"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const path_1 = require("path");
const fs_extra_1 = __importDefault(require("fs-extra"));
const memfs_1 = require("memfs");
function createPassiveFileSystem(caseSensitive = false) {
    // read cache
    const fsExistsCache = new Map();
    const fsReadStatsCache = new Map();
    const fsReadFileCache = new Map();
    const fsReadDirCache = new Map();
    const fsRealPathCache = new Map();
    function normalizePath(path) {
        return caseSensitive ? path_1.normalize(path) : path_1.normalize(path).toLowerCase();
    }
    // read methods
    function fsExists(path) {
        const normalizedPath = normalizePath(path);
        if (!fsExistsCache.has(normalizedPath)) {
            fsExistsCache.set(normalizedPath, fs_extra_1.default.existsSync(normalizedPath));
        }
        return !!fsExistsCache.get(normalizedPath);
    }
    function memExists(path) {
        return memfs_1.fs.existsSync(normalizePath(path));
    }
    function fsReadStats(path) {
        const normalizedPath = normalizePath(path);
        if (!fsReadStatsCache.has(normalizedPath)) {
            if (fsExists(normalizedPath)) {
                fsReadStatsCache.set(normalizedPath, fs_extra_1.default.statSync(normalizedPath));
            }
        }
        return fsReadStatsCache.get(normalizedPath);
    }
    function memReadStats(path) {
        return memExists(path) ? memfs_1.fs.statSync(normalizePath(path)) : undefined;
    }
    function fsReadFile(path, encoding) {
        const normalizedPath = normalizePath(path);
        if (!fsReadFileCache.has(normalizedPath)) {
            if (fsExists(normalizedPath)) {
                fsReadFileCache.set(normalizedPath, fs_extra_1.default.readFileSync(normalizedPath, { encoding }).toString());
            }
            else {
                fsReadFileCache.set(normalizedPath, undefined);
            }
        }
        return fsReadFileCache.get(normalizedPath);
    }
    function memReadFile(path, encoding) {
        if (memExists(path)) {
            return memfs_1.fs
                .readFileSync(normalizePath(path), { encoding: encoding })
                .toString();
        }
    }
    function fsReadDir(path) {
        const normalizedPath = normalizePath(path);
        if (!fsReadDirCache.has(normalizedPath)) {
            if (fsExists(normalizedPath)) {
                fsReadDirCache.set(normalizedPath, fs_extra_1.default.readdirSync(normalizedPath, { withFileTypes: true }));
            }
            else {
                fsReadDirCache.set(normalizedPath, []);
            }
        }
        return fsReadDirCache.get(normalizedPath) || [];
    }
    function memReadDir(path) {
        if (memExists(path)) {
            return memfs_1.fs.readdirSync(normalizePath(path), { withFileTypes: true });
        }
        return [];
    }
    function exists(path) {
        return fsExists(path) || memExists(path);
    }
    function readFile(path, encoding) {
        const fsStats = fsReadStats(path);
        const memStats = memReadStats(path);
        if (fsStats && memStats) {
            return fsStats.mtimeMs > memStats.mtimeMs
                ? fsReadFile(path, encoding)
                : memReadFile(path, encoding);
        }
        else if (fsStats) {
            return fsReadFile(path, encoding);
        }
        else if (memStats) {
            return memReadFile(path, encoding);
        }
    }
    function readDir(path) {
        const fsDirents = fsReadDir(path);
        const memDirents = memReadDir(path);
        // merge list of dirents from fs and mem
        return fsDirents
            .filter((fsDirent) => !memDirents.some((memDirent) => memDirent.name === fsDirent.name))
            .concat(memDirents);
    }
    function readStats(path) {
        const fsStats = fsReadStats(path);
        const memStats = memReadStats(path);
        if (fsStats && memStats) {
            return fsStats.mtimeMs > memStats.mtimeMs ? fsStats : memStats;
        }
        else if (fsStats) {
            return fsStats;
        }
        else if (memStats) {
            return memStats;
        }
    }
    function getRealPath(path) {
        const normalizedPath = normalizePath(path);
        if (!fsRealPathCache.has(normalizedPath)) {
            let base = normalizedPath;
            let nested = '';
            while (base !== path_1.dirname(base)) {
                if (fsExists(base)) {
                    fsRealPathCache.set(normalizedPath, normalizePath(path_1.join(fs_extra_1.default.realpathSync(base), nested)));
                    break;
                }
                nested = path_1.join(path_1.basename(base), nested);
                base = path_1.dirname(base);
            }
        }
        return fsRealPathCache.get(normalizedPath) || normalizedPath;
    }
    function createDir(path) {
        memfs_1.fs.mkdirSync(normalizePath(path), { recursive: true });
    }
    function writeFile(path, data) {
        if (!memExists(path_1.dirname(path))) {
            createDir(path_1.dirname(path));
        }
        memfs_1.fs.writeFileSync(normalizePath(path), data);
    }
    function deleteFile(path) {
        if (memExists(path)) {
            memfs_1.fs.unlinkSync(normalizePath(path));
        }
    }
    function updateTimes(path, atime, mtime) {
        if (memExists(path)) {
            memfs_1.fs.utimesSync(normalizePath(path), atime, mtime);
        }
    }
    return {
        exists(path) {
            return exists(getRealPath(path));
        },
        readFile(path, encoding) {
            return readFile(getRealPath(path), encoding);
        },
        readDir(path) {
            return readDir(getRealPath(path));
        },
        readStats(path) {
            return readStats(getRealPath(path));
        },
        realPath(path) {
            return getRealPath(path);
        },
        normalizePath(path) {
            return normalizePath(path);
        },
        writeFile(path, data) {
            writeFile(getRealPath(path), data);
        },
        deleteFile(path) {
            deleteFile(getRealPath(path));
        },
        createDir(path) {
            createDir(getRealPath(path));
        },
        updateTimes(path, atime, mtime) {
            updateTimes(getRealPath(path), atime, mtime);
        },
        clearCache() {
            fsExistsCache.clear();
            fsReadStatsCache.clear();
            fsReadFileCache.clear();
            fsReadDirCache.clear();
            fsRealPathCache.clear();
        },
    };
}
exports.createPassiveFileSystem = createPassiveFileSystem;
